/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <limits.h>

#include "xmap.h"

#define MAXORDER 3

typedef double DOUBLE;

/* structure for use internally with these functions.  these functions expect
   square matrices so only one variable is given (n) for the matrix size */

struct MATRIX {
    int n;			/* size of this matrix (n x n) */
    DOUBLE *v;
};

/* calculate offset into array based on r/c */

#define M(row,col) m->v[(((row)-1)*(m->n))+(col)-1]

/***************************************************************************/
/*
 */
/***************************************************************************/

#define MSUCCESS     1		/* success */
#define MNPTERR      0		/* not enough points */
#define MUNSOLVABLE -1		/* not solvable */
#define MMEMERR     -2		/* not enough memory */
#define MPARMERR    -3		/* parameter error */
#define MINTERR     -4		/* internal error */

static int calccoef();
static int calcls();
static int exactdet();
static int solvemat();
static DOUBLE term();



/*
   transform a single coordinate pair.
 */

int georef(double e1, double n1, double *e, double *n, double E[], double N[], int order)
{
    DOUBLE e3, e2n, en2, n3, e2, en, n2;

    switch (order) {
    case 1:

	*e = E[0] + E[1] * e1 + E[2] * n1;
	*n = N[0] + N[1] * e1 + N[2] * n1;
	break;

    case 2:

	e2 = e1 * e1;
	n2 = n1 * n1;
	en = e1 * n1;

	*e = E[0] + E[1] * e1 + E[2] * n1 +
	    E[3] * e2 + E[4] * en + E[5] * n2;
	*n = N[0] + N[1] * e1 + N[2] * n1 +
	    N[3] * e2 + N[4] * en + N[5] * n2;
	break;

    case 3:

	e2 = e1 * e1;
	en = e1 * n1;
	n2 = n1 * n1;
	e3 = e1 * e2;
	e2n = e2 * n1;
	en2 = e1 * n2;
	n3 = n1 * n2;

	*e = E[0] +
	    E[1] * e1 + E[2] * n1 +
	    E[3] * e2 + E[4] * en + E[5] * n2 +
	    E[6] * e3 + E[7] * e2n + E[8] * en2 + E[9] * n3;
	*n = N[0] +
	    N[1] * e1 + N[2] * n1 +
	    N[3] * e2 + N[4] * en + N[5] * n2 +
	    N[6] * e3 + N[7] * e2n + N[8] * en2 + N[9] * n3;
	break;

    default:

	return (MPARMERR);
	break;
    }

    return (MSUCCESS);
}

/***************************************************************************/
/*
   compute the georefferencing coefficients based on a set of control points
 */
/***************************************************************************/

int compute_georef_equations(struct GeoRef *cp, double E12[], double N12[],
					double E21[], double N21[])
{
    double *tempptr;
    int status;

    if (cp->order < 1 || cp->order > MAXORDER)
	return (MPARMERR);

    /* calculate the forward transformation coefficients */

    status = calccoef(cp, E12, N12, cp->order);
    if (status != MSUCCESS)
	return (status);

    /* switch the 1 and 2 easting and northing arrays */

    tempptr = cp->e1;
    cp->e1 = cp->e2;
    cp->e2 = tempptr;
    tempptr = cp->n1;
    cp->n1 = cp->n2;
    cp->n2 = tempptr;

    /* calculate the backward transformation coefficients */

    status = calccoef(cp, E21, N21, cp->order);

    /* switch the 1 and 2 easting and northing arrays back */

    tempptr = cp->e1;
    cp->e1 = cp->e2;
    cp->e2 = tempptr;
    tempptr = cp->n1;
    cp->n1 = cp->n2;
    cp->n2 = tempptr;

    return (status);
}

/***************************************************************************/
/*
   compute the georefferencing coefficients based on a set of control points
 */
/***************************************************************************/

static int calccoef(cp, E, N, order)
struct GeoRef *cp;
double E[];
double N[];
int order;
{
    struct MATRIX m;
    DOUBLE *a;
    DOUBLE *b;
    int numactive;		/* number of active control points */
    int status, i;

    /* calculate the number of valid control points */

    numactive = cp->count;

    /* calculate the minimum number of control points needed to determine
       a transformation of this order */

    m.n = ((order + 1) * (order + 2)) / 2;

    if (numactive < m.n)
	return (MNPTERR);

    /* INITIALIZE MATRIX */

    m.v = (DOUBLE *) calloc(m.n * m.n, sizeof(DOUBLE));
    if (m.v == NULL) {
	return (MMEMERR);
    }
    a = (DOUBLE *) calloc(m.n, sizeof(DOUBLE));
    if (a == NULL) {
	free((char *) m.v);
	return (MMEMERR);
    }
    b = (DOUBLE *) calloc(m.n, sizeof(DOUBLE));
    if (b == NULL) {
	free((char *) m.v);
	free((char *) a);
	return (MMEMERR);
    }
    if (numactive == m.n)
	status = exactdet(cp, &m, a, b, E, N);
    else
	status = calcls(cp, &m, a, b, E, N);

    free((char *) m.v);
    free((char *) a);
    free((char *) b);

    return (status);
}

/***************************************************************************/
/*
   calculate the transformation coefficients with exactly the minimum
   number of control points required for this transformation.
 */
/***************************************************************************/

static int exactdet(cp, m, a, b, E, N)
struct GeoRef *cp;
struct MATRIX *m;
DOUBLE a[];
DOUBLE b[];
double E[];			/* easting coefficients */
double N[];			/* northing coefficients */
{
    int pntnow, currow, j;

    currow = 1;
    for (pntnow = 0; pntnow < cp->count; pntnow++) {
	    /* POPULATE MATRIX M */

	    for (j = 1; j <= m->n; j++) {
		M(currow, j) = term(j, cp->e1[pntnow], cp->n1[pntnow]);
	    }

	    /* populate matrix a and b */

	    a[currow - 1] = cp->e2[pntnow];
	    b[currow - 1] = cp->n2[pntnow];

	    currow++;
    }

    if (currow - 1 != m->n)
	return (MINTERR);

    return (solvemat(m, a, b, E, N));
}

/***************************************************************************/
/*
   calculate the transformation coefficients with more than the minimum
   number of control points required for this transformation.  this
   routine uses the least squares method to compute the coefficients.
 */
/***************************************************************************/

static int calcls(cp, m, a, b, E, N)
struct GeoRef *cp;
struct MATRIX *m;
DOUBLE a[];
DOUBLE b[];
double E[];			/* EASTING COEFFICIENTS */
double N[];			/* NORTHING COEFFICIENTS */
{
    int i, j, n, numactive = 0;

    /* initialize the upper half of the matrix and the two column vectors */

    for (i = 1; i <= m->n; i++) {
	for (j = i; j <= m->n; j++)
	    M(i, j) = 0.0;
	a[i - 1] = b[i - 1] = 0.0;
    }

    /* sum the upper half of the matrix and the column vectors according to
       the least squares method of solving over determined systems */

    for (n = 0; n < cp->count; n++) {
	    numactive++;
	    for (i = 1; i <= m->n; i++) {
		for (j = i; j <= m->n; j++)
		    M(i, j) += term(i, cp->e1[n], cp->n1[n]) * term(j, cp->e1[n], cp->n1[n]);

		a[i - 1] += cp->e2[n] * term(i, cp->e1[n], cp->n1[n]);
		b[i - 1] += cp->n2[n] * term(i, cp->e1[n], cp->n1[n]);
	    }
    }

    if (numactive <= m->n)
	return (MINTERR);

    /* transpose values in upper half of m to other half */

    for (i = 2; i <= m->n; i++) {
	for (j = 1; j < i; j++)
	    M(i, j) = M(j, i);
    }

    return (solvemat(m, a, b, E, N));
}

/***************************************************************************/
/*
   calculate the x/y term based on the term number

   ORDER\TERM   1    2    3    4    5    6    7    8    9   10
   1        e0n0 e1n0 e0n1
   2        e0n0 e1n0 e0n1 e2n0 e1n1 e0n2
   3        e0n0 e1n0 e0n1 e2n0 e1n1 e0n2 e3n0 e2n1 e1n2 e0n3
 */
/***************************************************************************/

static DOUBLE term(term, e, n)
int term;
double e;
double n;
{
    switch (term) {
    case 1:
	return ((DOUBLE) 1.0);
    case 2:
	return ((DOUBLE) e);
    case 3:
	return ((DOUBLE) n);
    case 4:
	return ((DOUBLE) (e * e));
    case 5:
	return ((DOUBLE) (e * n));
    case 6:
	return ((DOUBLE) (n * n));
    case 7:
	return ((DOUBLE) (e * e * e));
    case 8:
	return ((DOUBLE) (e * e * n));
    case 9:
	return ((DOUBLE) (e * n * n));
    case 10:
	return ((DOUBLE) (n * n * n));
    }
    return ((DOUBLE) 0.0);
}

/***************************************************************************/
/*
   solve for the 'E' and 'N' coefficients by using a somewhat modified
   gaussian elimination method.

   | M11 M12 ... M1n | | E0   |   | a0   |
   | M21 M22 ... M2n | | E1   | = | a1   |
   |  .   .   .   .  | | .    |   | .    |
   | Mn1 Mn2 ... Mnn | | En-1 |   | an-1 |

   and

   | M11 M12 ... M1n | | N0   |   | b0   |
   | M21 M22 ... M2n | | N1   | = | b1   |
   |  .   .   .   .  | | .    |   | .    |
   | Mn1 Mn2 ... Mnn | | Nn-1 |   | bn-1 |
 */
/***************************************************************************/

static int solvemat(m, a, b, E, N)
struct MATRIX *m;
DOUBLE a[];
DOUBLE b[];
double E[];
double N[];
{
    int i, j, i2, j2, imark;
    DOUBLE factor, temp;
    DOUBLE pivot;		/* actual value of the largest pivot candidate */

    for (i = 1; i <= m->n; i++) {
	j = i;

	/* find row with largest magnitude value for pivot value */

	pivot = M(i, j);
	imark = i;
	for (i2 = i + 1; i2 <= m->n; i2++) {
	    temp = fabs(M(i2, j));
	    if (temp > fabs(pivot)) {
		pivot = M(i2, j);
		imark = i2;
	    }
	}

	/* if the pivot is very small then the points are nearly co-linear */
	/* co-linear points result in an undefined matrix, and nearly */
	/* co-linear points results in a solution with rounding error */

	if (pivot == 0.0)
	    return (MUNSOLVABLE);

	/* if row with highest pivot is not the current row, switch them */

	if (imark != i) {
	    for (j2 = 1; j2 <= m->n; j2++) {
		temp = M(imark, j2);
		M(imark, j2) = M(i, j2);
		M(i, j2) = temp;
	    }

	    temp = a[imark - 1];
	    a[imark - 1] = a[i - 1];
	    a[i - 1] = temp;

	    temp = b[imark - 1];
	    b[imark - 1] = b[i - 1];
	    b[i - 1] = temp;
	}
	/* compute zeros above and below the pivot, and compute
	   values for the rest of the row as well */

	for (i2 = 1; i2 <= m->n; i2++) {
	    if (i2 != i) {
		factor = M(i2, j) / pivot;
		for (j2 = j; j2 <= m->n; j2++)
		    M(i2, j2) -= factor * M(i, j2);
		a[i2 - 1] -= factor * a[i - 1];
		b[i2 - 1] -= factor * b[i - 1];
	    }
	}
    }

    /* since all other values in the matrix are zero now, calculate the
       coefficients by dividing the column vectors by the diagonal values. */

    for (i = 1; i <= m->n; i++) {
	E[i - 1] = a[i - 1] / M(i, i);
	N[i - 1] = b[i - 1] / M(i, i);
    }

    return (MSUCCESS);
}

